@@ -9,36 +9,41 @@ import jax.numpy as jnp
 
 
 ###############################################################################
-# Context
+# State
 ###############################################################################
 
 
 # When running an impure function that accesses state, it will find the state
 # in this global variable. The pure() wrapper populates this global variable
-# with the provided state, calles the inner function, and the takes the
-# resulting state out of the global variable and returns it back to the user.
+# with the provided state, calls the inner function, and then the takes the
+# resulting state out of the global variable to return it back to the user.
 CONTEXT = [None]
 
 
-class Context:
+class Context(dict):
 
-  def __init__(self, state, rng, create, modify, freeze, reserve):
-    self.state = state
-    self.rng = rng
+  def __init__(self, entries, rng, create, modify, ignore, reserve):
+    super().__init__(entries)
     self.create = create  # Allow creating new state entries.
-    self.modify = modify  # Apply modifications to existing state entries.
-    self.freeze = freeze  # Disllow modifying existing state entries.
+    self.modify = modify  # Allow modifying existing state entries.
+    self.ignore = ignore  # Ignore modifications to existing state entries.
+    self.rng = rng
     self.reserve = reserve
 
+  def update(self, entries):
+    for key, value in dict(entries).items():
+      self[key] = value
 
-def context():
-  """Access and modify the global context from within an impure function. For
-  advanced users only. Prefer to use state() to access and modify the global
-  state and rng() to get the next RNG key."""
-  global CONTEXT
-  if CONTEXT[0] is None:
-    raise RuntimeError('Wrap impure functions in pure() before running them.')
-  return CONTEXT[0]
+  def __setitem__(self, key, value):
+    if not self.modify:
+      raise RuntimeError(
+          'Cannot modify state entries here. If you want to modify '
+          'state inside of scan() set modify=True.')
+    if self.ignore and key in self:
+      return  # Do not overwrite existing entries.
+    if not self.create and key not in self:
+      raise RuntimeError('Can only create state entries during first call.')
+    super().__setitem__(key, value)
 
 
 def pure(fun, nested=False):
@@ -47,17 +52,28 @@ def pure(fun, nested=False):
   transformation. The pure function can be used as follows:
   `out, state = fun(state, rng, *args, **kwargs)`."""
   def purified(
-      state, rng, *args, create=True, modify=True, freeze=False, **kwargs):
+      state, rng, *args, create=None, modify=None, ignore=None, **kwargs):
     global CONTEXT
+    if CONTEXT[0]:
+      create = create if create is not None else CONTEXT[0].create
+      modify = modify if modify is not None else CONTEXT[0].modify
+      ignore = ignore if ignore is not None else CONTEXT[0].ignore
+      assert CONTEXT[0].create or not create, 'Parent context disabled create.'
+      assert CONTEXT[0].modify or not modify, 'Parent context disabled modify.'
+      assert not CONTEXT[0].ignore or ignore, 'Parent context enabled ignore.'
+    else:
+      create = create if create is not None else True
+      modify = modify if modify is not None else True
+      ignore = ignore if ignore is not None else False
     if not isinstance(state, dict):
       raise ValueError('Must provide a dict as state.')
     if CONTEXT[0] and (not nested):
       raise RuntimeError('If you want to nest run() calls, use nested=True.')
     before = CONTEXT[0]
     try:
-      CONTEXT[0] = Context(state.copy(), rng, create, modify, freeze, [])
+      CONTEXT[0] = Context(state.copy(), rng, create, modify, ignore, [])
       out = fun(*args, **kwargs)
-      state = CONTEXT[0].state
+      state = dict(CONTEXT[0])
       return out, state
     finally:
       CONTEXT[0] = before
@@ -65,11 +81,14 @@ def pure(fun, nested=False):
   return purified
 
 
-def state():
-  """Access or modify the global state dictionary."""
-  # TODO: Wrap state dictionary with guards for context().create and
-  # context().modify.
-  return context().state
+def context():
+  """Access and modify the global context from within an impure function. For
+  advanced users only. Prefer to use module methods to access and modify state
+  and rng() to get the next RNG key."""
+  global CONTEXT
+  if CONTEXT[0] is None:
+    raise RuntimeError('Wrap impure functions in pure() before running them.')
+  return CONTEXT[0]
 
 
 def rng(amount=None, reserve=16):
@@ -88,9 +107,9 @@ def rng(amount=None, reserve=16):
 
 
 def creating():
-  """Boolean indicating whether the program is currently allowed to create
-  state entries. Can use used for initialization logic that should be excluded
-  from compiled functions."""
+  """Indicates whether the program is currently allowed to create state
+  entries. Can use used for initialization logic that should be excluded from
+  compiled functions."""
   return context().create
 
 
@@ -111,26 +130,21 @@ def grad(fun, keys, has_aux=False):
     fun = lambda *args, _fun=fun, **kwargs: (_fun(*args, *kwargs), {})
   fun = pure(fun, nested=True)
   def forward(x1, x2, rng, *args, **kwargs):
-    (y, aux), state = fun(
-        {**x1, **x2}, rng, *args, create=False, modify=True, **kwargs)
+    (y, aux), state = fun({**x1, **x2}, rng, *args, create=False, **kwargs)
     return y, (aux, state)
   backward = jax.value_and_grad(forward, has_aux=True)
   @functools.wraps(backward)
   def wrapper(*args, **kwargs):
-    ctx = context()
-    if ctx.create:
-      _, state = fun(
-          ctx.state, rng(), *args, create=True, modify=False, **kwargs)
-      ctx.state.update(state)
+    prerun(fun, *args, **kwargs)
     assert all(isinstance(x, (str, Module)) for x in keys)
     strs = [x for x in keys if isinstance(x, str)]
     mods = [x for x in keys if isinstance(x, Module)]
     for mod in mods:
       strs += mod.getm()
-    x1 = {k: v for k, v in ctx.state.items() if k in strs}
-    x2 = {k: v for k, v in ctx.state.items() if k not in strs}
+    x1 = {k: v for k, v in context().items() if k in strs}
+    x2 = {k: v for k, v in context().items() if k not in strs}
     (y, (aux, state)), dx = backward(x1, x2, rng(), *args, **kwargs)
-    ctx.state.update(state)
+    context().update(state)
     return (y, x1, dx, aux) if has_aux else (y, x1, dx)
   return wrapper
 
@@ -141,78 +155,88 @@ def jit(fun, static=None, **kwargs):
   if not getattr(fun, 'pure', False):
     raise ValueError('Use pure() before applying jit().')
   static = static or ()
-  compiled = jax.jit(
-      lambda sta, *a, **kw: fun(*a, **kw, **dict(sta), create=False),
-      static_argnums=[0], **kwargs)
-  @functools.wraps(compiled)
-  def wrapper(*args, **kwargs):
-    for name in static:
-      if name not in kwargs:
-        raise ValueError('Please pass all static arguments by keyword.')
-    if not wrapper.created:
-      wrapper.created = True
-      return fun(*args, create=True, **kwargs)
-    nosta = {k: v for k, v in kwargs.items() if k not in sta}
-    sta = {k: v for k, v in kwargs.items() if k in static}
-    sta = tuple(sorted(sta.items()))
-    return compiled(sta, *args, **nosta)
-  wrapper.created = False
+
+  @bind(jax.jit, static_argnums=[0], **kwargs)
+  def init(statics, rng, *args, **kwargs):
+    # Return only state so JIT can remove dead code for fast initialization.
+    s = fun({}, rng, *args, ignore=True, **kwargs)[1]
+    return s
+
+  @bind(jax.jit, static_argnums=[0], **kwargs)
+  def apply(statics, state, rng, *args, **kwargs):
+    return fun(state, rng, *args, create=False, **dict(statics), **kwargs)
+
+  @functools.wraps(fun)
+  def wrapper(state, rng, *args, **kwargs):
+    if any([name not in kwargs for name in static]):
+      raise ValueError('Please pass all static arguments by keyword.')
+    state = state.copy()
+    statics = tuple(sorted([(k, v) for k, v in kwargs.items() if k in static]))
+    kwargs = {k: v for k, v in kwargs.items() if k not in static}
+    if not hasattr(wrapper, 'keys'):
+      created = init(statics, rng, *args, **kwargs)
+      wrapper.keys = set(created.keys())
+      for key, value in created.items():
+        if key not in state:
+          state[key] = value
+    selected = {k: v for k, v in state.items() if k in wrapper.keys}
+    out, updated = apply(statics, selected, rng, *args, **kwargs)
+    return out, {**state, **updated}
   return wrapper
 
 
-def pmap(fun, axis_name=None, static=None, in_axes=0, out_axes=0, **kwargs):
+def pmap(fun, axis_name=None, static=None, **kwargs):
   """Compiles n pure function for fast execution across multiple devices. Only
   the first call of the function is allowed to create state entries."""
-  static = static or ()
   if not getattr(fun, 'pure', False):
-    raise ValueError('Use pure() before applying pmap().')
-  inner = lambda sta, *a, **kw: fun(*a, **kw, **dict(sta), create=False)
-  compiled = jax.pmap(
-      inner, axis_name, in_axes=in_axes, out_axes=out_axes,
-      static_broadcasted_argnums=[0], **kwargs)
-  @functools.wraps(compiled)
-  def wrapper(*args, **kwargs):
-    for name in static:
-      if name not in kwargs:
-        raise ValueError('Please pass all static arguments by keyword.')
-    nosta = {k: v for k, v in kwargs.items() if k not in static}
-    sta = {k: v for k, v in kwargs.items() if k in static}
-    sta = tuple(sorted(sta.items()))
-    if not wrapper.created:
-      wrapper.created = True
-      return jax.vmap(
-          functools.partial(inner, sta),
-          in_axes, out_axes, axis_name)(*args, **nosta)
-    return compiled(sta, *args, **nosta)
-  wrapper.created = False
+    raise ValueError('Use pure() before applying jit().')
+  static = static or ()
+
+  @bind(jax.pmap, axis_name=axis_name, static_broadcasted_argnums=[0], **kwargs)
+  def init(statics, rng, *args, **kwargs):
+    # Return only state so JIT can remove dead code for fast initialization.
+    return fun({}, rng, *args, ignore=True, **kwargs)[1]
+
+  @bind(jax.pmap, axis_name=axis_name, static_broadcasted_argnums=[0], **kwargs)
+  def apply(statics, state, rng, *args, **kwargs):
+    return fun(state, rng, *args, create=False, **dict(statics), **kwargs)
+
+  @functools.wraps(fun)
+  def wrapper(state, rng, *args, **kwargs):
+    if any([name not in kwargs for name in static]):
+      raise ValueError('Please pass all static arguments by keyword.')
+    state = state.copy()
+    statics = tuple(sorted([(k, v) for k, v in kwargs.items() if k in static]))
+    kwargs = {k: v for k, v in kwargs.items() if k not in static}
+    if not hasattr(wrapper, 'keys'):
+      created = init(statics, rng, *args, **kwargs)
+      wrapper.keys = set(created.keys())
+      for key, value in created.items():
+        if key not in state:
+          state[key] = value
+    selected = {k: v for k, v in state.items() if k in wrapper.keys}
+    out, updated = apply(statics, selected, rng, *args, **kwargs)
+    return out, {**state, **updated}
   return wrapper
 
 
 def cond(pred, true_fun, false_fun, *operands):
-  ctx = context()
   true_fun = pure(true_fun, nested=True)
   false_fun = pure(false_fun, nested=True)
-  if ctx.create:
-    _, state = true_fun(ctx.state, rng(), *operands, create=True, modify=False)
-    ctx.state.update(state)
-    _, state = false_fun(ctx.state, rng(), *operands, create=True, modify=False)
-    ctx.state.update(state)
+  prerun(true_fun, *operands)
+  prerun(false_fun, *operands)
   out, state = jax.lax.cond(
       pred,
       lambda state, rng1, rng2, *args: true_fun(state, rng1, *args),
       lambda state, rng1, rng2, *args: false_fun(state, rng2, *args),
-      ctx.state, *rng(2), *operands)
-  ctx.state.update(state)
+      dict(context()), *rng(2), *operands)
+  context().update(state)
   return out
 
 
 def scan(fun, carry, xs, reverse=False, unroll=1, modify=False):
-  ctx = context()
   fun = pure(fun, nested=True)
-  if ctx.create:
-    first = jax.tree_util.tree_map(lambda x: x[0], xs)
-    _, state = fun(ctx.state, rng(), carry, first, create=True, modify=False)
-    ctx.state.update(state)
+  prerun(fun, carry, jax.tree_util.tree_map(lambda x: x[0], xs))
   length = len(jax.tree_util.tree_leaves(xs)[0])
   rngs = rng(length)
   if modify:
@@ -222,18 +246,28 @@ def scan(fun, carry, xs, reverse=False, unroll=1, modify=False):
       (carry, y), state = fun(state, rng, carry, x, create=False)
       return (carry, state), y
     (carry, state), ys = jax.lax.scan(
-        inner, (carry, ctx.state), (xs, rngs), length, reverse, unroll)
-    ctx.state.update(state)
+        inner, (carry, dict(context())), (xs, rngs), length, reverse, unroll)
+    context().update(state)
   else:
     def inner(carry, x):
       x, rng = x
       (carry, y), state = fun(
-          ctx.state, rng, carry, x, create=False, freeze=True)
+          dict(context()), rng, carry, x, create=False, modify=False)
       return carry, y
     carry, ys = jax.lax.scan(inner, carry, (xs, rngs), length, reverse, unroll)
   return carry, ys
 
 
+def prerun(fun, *args, **kwargs):
+  if not context().create:
+    return
+  discarded, state = fun(dict(context()), rng(), *args, ignore=True, **kwargs)
+  jax.tree_util.tree_map(
+      lambda x: hasattr(x, 'delete') and x.delete(), discarded)
+  context().update(state)
+
+
+
 ###############################################################################
 # Modules
 ###############################################################################
@@ -335,19 +369,18 @@ class Module(object, metaclass=ModuleMeta):
 
   def get(self, name, *args, **kwargs):
     """Retrieve or create a state entry that belongs to this module."""
-    state_ = state()
     path = self.path + '/' + name
     if name in self._submodules:
       return self._submodules[name]
-    if path in state_:
-      return state_[path]
+    if path in context():
+      return context()[path]
     ctor, *args = args
     if 'name' in inspect.signature(ctor).parameters:
       kwargs['name'] = name
     value = ctor(*args, **kwargs)
     flat, _ = jax.tree_util.tree_flatten(value)
     if all(isinstance(x, jnp.ndarray) for x in flat):
-      state_[path] = value
+      context()[path] = value
     else:
       self._submodules[name] = value
     return value
@@ -359,11 +392,10 @@ class Module(object, metaclass=ModuleMeta):
 
   def getm(self, pattern=r'.*', allow_empty=False):
     """Read the state entries of this module, optionally filtered by regex."""
-    state_ = state()
     pattern = re.compile(pattern)
     prefix = self.path + '/'
     results = {}
-    for key, value in state_.items():
+    for key, value in context().items():
       if not key.startswith(prefix):
         continue
       if pattern.match(key[len(prefix):]):
@@ -378,19 +410,7 @@ class Module(object, metaclass=ModuleMeta):
     for key in mapping:
       if not key.startswith(prefix):
         raise KeyError(f'Key {key} does not belong to module {self.path}.')
-    ctx = context()
-    if ctx.freeze:
-      raise RuntimeError(
-          'Cannot modify state entries here. If you want to modify '
-          'state inside of scan() set modify=True.')
-    if not ctx.modify:
-      mapping = {k: v for k, v in mapping.items() if k not in ctx.state}
-    if not ctx.create:
-      for key, value in mapping.items():
-        if key not in ctx.state:
-          raise RuntimeError(
-              f'Can only create state entries during first call ({key}).')
-    ctx.state.update(mapping)
+    context().update(mapping)
 
 
 class Variable(Module):
@@ -446,5 +466,5 @@ class OptaxModule(Module):
     optstate = self.get('state', self.opt.init, params)
     updates, optstate = self.opt.update(grads, optstate)
     self.put('state', optstate)
-    state().update(optax.apply_updates(params, updates))
+    context().update(optax.apply_updates(params, updates))
     return {'loss': loss.mean(), 'grad_norm': optax.global_norm(grads)}
